module adj_split_w2_field_kernel_mod
  use argument_mod, only : any_space_1, arg_type, cell_column, gh_field, gh_inc, gh_read, gh_readwrite, gh_real
  use constants_mod, only : i_def, r_def
  use fs_continuity_mod, only : w2, w2h, w2v
  use kernel_mod, only : kernel_type
  implicit none
  type, public, extends(kernel_type) :: adj_split_w2_field_kernel_type
  type(ARG_TYPE) :: META_ARGS(3) = (/ &
    arg_type(gh_field, gh_real, gh_inc, w2h), &
    arg_type(gh_field, gh_real, gh_readwrite, w2v), &
    arg_type(gh_field, gh_real, gh_inc, w2)/)
  INTEGER :: OPERATES_ON = cell_column
  CONTAINS
    PROCEDURE, NOPASS :: adj_split_w2_field_code
END TYPE adj_split_w2_field_kernel_type

  private

  public :: adj_split_w2_field_code

  contains
  subroutine adj_split_w2_field_code(nlayers, uv, w, uvw, ndf_w2h, undf_w2h, map_w2h, ndf_w2v, undf_w2v, map_w2v, ndf_w2, undf_w2, &
&map_w2)
    integer(kind=i_def), intent(in) :: nlayers
    integer(kind=i_def), intent(in) :: ndf_w2
    integer(kind=i_def), intent(in) :: ndf_w2h
    integer(kind=i_def), intent(in) :: ndf_w2v
    integer(kind=i_def), intent(in) :: undf_w2
    integer(kind=i_def), intent(in) :: undf_w2h
    integer(kind=i_def), intent(in) :: undf_w2v
    integer(kind=i_def), dimension(ndf_w2), intent(in) :: map_w2
    integer(kind=i_def), dimension(ndf_w2h), intent(in) :: map_w2h
    integer(kind=i_def), dimension(ndf_w2v), intent(in) :: map_w2v
    real(kind=r_def), dimension(undf_w2h), intent(inout) :: uv
    real(kind=r_def), dimension(undf_w2v), intent(inout) :: w
    real(kind=r_def), dimension(undf_w2), intent(inout) :: uvw
    integer(kind=i_def) :: df
    integer(kind=i_def) :: k

    do k = nlayers - 1, 0, -1
      do df = ndf_w2v, 1, -1
        uvw(map_w2(ndf_w2h + df) + k) = uvw(map_w2(ndf_w2h + df) + k) + w(map_w2v(df) + k)
        w(map_w2v(df) + k) = 0.0
      enddo
      do df = ndf_w2h, 1, -1
        uvw(map_w2(df) + k) = uvw(map_w2(df) + k) + uv(map_w2h(df) + k)
        uv(map_w2h(df) + k) = 0.0
      enddo
    enddo

  end subroutine adj_split_w2_field_code

end module adj_split_w2_field_kernel_mod
